# System dynamics modeling scripts

# Code from internet discussion group somewhere
library(deSolve)

##########################################
# Solving systems of ODEs
##########################################

# Construct matrix from inferelator model
# The matrix has cluster influences as the columns and
#    cluster targets as the rows. A cluster influence is the 
#    cluster that contains a regulator that is inferred to
#    have an influence on the cluster target. An alternative
#    is to use the clusters themselves as influences and 
#    short circuit this process.
# Note: copied from InferelatorUtilities.R

regulator_matrix <- function(model_list, score_filter=NA, regulators_in=NA, addrownames=TRUE) {
	# takes a model (list of clusters, each a list of their regulatory influences)
	#       and creates a matrix from this. Rows are targets and columns are influences.
	
	if (!is.na(regulators_in)) {
		regulators = regulators_in
	} else {
		regulators = c()
	
		# get a list of regulatory influences from the list of models
		for (i in 1:length(model_list)) {
			for (j in 1:length(model_list[[i]])) {
				n = names(model_list[[i]][j])[1]
				x = model_list[[i]][j]
				if (!is.na(score_filter)) {
					if (abs(x) > score_filter) {
						if (!n %in% regulators) regulators = c(regulators, n)
					}
				} else {
					if (!n %in% regulators) regulators = c(regulators, n)
				}
			}
		}
	}
	
	#outmat = matrix(nrow=length(model_list)+2, ncol=length(regulators))
	outmat = matrix(data=rep(0, length(model_list)*length(regulators)), nrow=length(model_list), ncol=length(regulators))
	colnames(outmat) = regulators
	# this assumes that the list of models has names
	#rownames(outmat) = c(names(model_list), "count", "meanweight")
	rownames(outmat) = names(model_list)
	
	if (addrownames) {
		rownames(outmat) = sapply(1:nrow(outmat), function (i) paste("cluster_", i, sep=""))
	}
	
	#outmat["count",] = rep(0,length(regulators))
	
	#cat(dim(outmat), "\n")
	for (i in 1:length(model_list)) {
		for (j in 1:length(model_list[[i]])) {
			n = names(model_list[[i]][j])[1]
			x = model_list[[i]][j]
			if (!is.na(score_filter)) {
				if (abs(x) > score_filter) {
					outmat[i,n] = x
					#outmat["count",n] = outmat["count",n]+1
				}
			} else {
				#cat(dim(outmat), i, n, "\n")
				outmat[i,n] = x
				#outmat["count",n] = outmat["count",n]+1
			}
		}
	}
	#outmat["meanweight",] = sapply(1:ncol(outmat), function (c) mean(outmat[1:length(model_list),c], na.rm=T))
	#outmat = outmat[,order(outmat["count",], decreasing=T)]
	return(outmat)	
}

regulator_matrices <- function(results_list, eigenvalues=F) {
	# this will work for results generated with the "self-prediction" option set to TRUE
	matrices = list()
	for (result in results_list) {
		model = result$model
		tfNames = sapply(1:length(model), function (i) paste("cluster_", i, sep=""))
		m = regulator_matrix(model, regulators_in=tfNames)
		rownames(m) = tfNames
		if (eigenvalues) {
			m = Re(eigen(m)$values)	
		}
		matrices = c(matrices, list(m))
	}
	return(matrices)
}

matrix_to_model <- function(inmat) {
	# converts an influence matrix to a model list
	model = list()
	for (col in colnames(inmat)) {
		influences = c()
		labels = c()
		for (row in rownames(inmat)) {
			if (abs(inmat[row,col])>0) {
				influences = c(influences, inmat[row,col])
				labels = c(labels, row)
			}
		}
		names(influences) = labels
		model = c(model, list(influences))
	}
	return(model)
}

matrix_to_edges <- function(inmat, edgefile="edges.out") {
	for (col in colnames(inmat)) {
		for (row in rownames(inmat)) {
			if (inmat[row,col] != 0) {
				cat(col, "\t", row, "\t", inmat[row,col], "\n", file=edgefile, append=TRUE)
			}
		}	
	}
}

random_matrix <- function(size, range=c(-2,2), labels=NA) {
	m = matrix(runif(size*size, range[1], range[2]), nrow=size, ncol=size)
	if (!is.na(labels)) {
		rownames(m) = labels
		colnames(m) = labels	
	}
	return(m)
}

random_matrices <- function(size, range=c(-2,2), ntimes=100, labels=NA) {
	maxevals = c()
	matrices = list()	
	for (i in 1:ntimes) {
		m = random_matrix(size, range, labels)	
		matrices = c(matrices, list(m))
		es = max(Re(eigen(m)$values))
		maxevals = c(maxevals, es)
	}
	
	return(list(maxevals=maxevals, matrices=matrices))
}

scramble_matrix <- function(inmat, mode="scramble", range=c(-2,2), mean=0, sd=1) {
	# returns a 'scambled' matrix. Options are as follows:
	#   mode="scramble" : randomly reorders the matrix, but keeps the same values
	#   mode="replace"  : replaces nonzero values with a random value from uniform distribution (given the input range)
	#   mode="replace-sample" : replaces nonzero values with a randomly selected value from the input matrix
	#   mode="replace-perturb": perturb nonzero values by an random amount amount proportional to their value
	#
	if (mode == "scramble") {	
		return(matrix(sample(inmat, length(inmat)), nrow=nrow(inmat), ncol=ncol(inmat), dimnames=dimnames(inmat)))
	}
	if (mode == "replace") {
		x = which(abs(inmat)>0)
		y = runif(length(x), range[1], range[2])
		outmat = inmat
		outmat[x] = y
		return(outmat)
	}
	if (mode == "replace-sample") {
		x = which(abs(inmat)>0)	
		y = sample(inmat[x], length(x))
		outmat = inmat
		outmat[x] = y
		return(outmat)
	}
	if (mode == "replace-perturb") {
		x = which(abs(inmat)>0)	
		y = rnorm(length(x), mean=mean, sd=sd)
		outmat = inmat
		outmat[x] = outmat[x] + outmat[x] * y
		return(outmat)
	}
}

scramble_matrices <- function(inmat, ntimes=100, mode="scramble", range=c(-2,2), mean=0, sd=1, collect=FALSE, calc_eigenvalues=FALSE, maxeval_thresh=0.5) {
	maxevals = c()
	matrices = list()
	count = 0
	for (i in 1:ntimes) {
		m = scramble_matrix(inmat, mode=mode, range=range, mean=mean, sd=sd)
		if (calc_eigenvalues) {
			es = max(Re(eigen(m)$values))
		}
		recflag = FALSE
		if (collect) {
			if (es < maxeval_thresh) {
				count = count + 1
				recflag = TRUE
				cat("Collected one at ", es, "\n")
				if (count > collect) {
					break	
				}
			}	
		} else {
			recflag = TRUE	
		}
		
		if (recflag == TRUE) {
			matrices = c(matrices, list(m))
			if (calc_eigenvalues) {
				maxevals = c(maxevals, es)
			}
		}
	}
	return(list(maxevals=maxevals, matrices=matrices))
}

# Combine/compare matrices in various ways to analyze ensembles of models
combine_matrices <- function(matrices=NA, optresults=NA, mode="occupancy", minocc=0.5) {
	if (is.na(matrices)) {
		# extract matrices from the optresults (optimization results) structure
		matrices = sapply(optresults, function (r) list(r[[1]]))
	}
	
	result = matrix(0, nrow=nrow(matrices[[1]]), ncol=ncol(matrices[[1]]))
	colnames(result) = colnames(matrices[[1]])
	rownames(result) = rownames(matrices[[1]])
	
	if (mode == "occupancy") {
		result = (result != 0)	
	}
	if (mode == "median") {
		result = array(0, dim=c(nrow(matrices[[1]]), ncol(matrices[[1]]), length(matrices)))
		i = 0
	}
	if (mode == "meanocc") {
		occ = (result != 0)	
	}
	
	for (m in matrices) {
		if (mode == "occupancy") {
			result = result + (m !=0)
		}
		else if (mode == "mean") {
			result = result + m	
		}
		else if (mode == "median") {
			i = i + 1
			result[,,i] = m
		}
		else if (mode == "meanocc") {
			occ = occ + (m != 0)
			result = result + m	
		}
	}
	
	if (mode == "mean") {
		result = result / length(matrices)	
	} 
	else if (mode == "median") {
		out = matrix(0, nrow=nrow(matrices[[1]]), ncol=ncol(matrices[[1]]))
		colnames(out) = colnames(matrices[[1]])
		rownames(out) = rownames(matrices[[1]])

		for (i in 1:nrow(out)) {
			for (j in 1:ncol(out)) {
				out[i,j] = median(result[i,j,])
			}	
		}
		result = out
	}
	else if (mode == "meanocc") {
		result = result / length(matrices)
		
		# filter mean results for occupancy
		result[which(occ < (length(matrices)*minocc))] = 0
	}
	
	
	return(result)
}

evaluate_matrices_conditions <- function(matrices, ratios=NA, counts=NA, times=NA, conditions=NA, observation_indices=NA,
									method="euler", eval.mode="correlation", include_combined_matrix=TRUE,
									return.results=FALSE, tau=NA) {
		# function to evaluate a set of models (matrices) on a set of conditions
		# it is assumed that the initial state of the system is the first column in each
		# condition set
		first = TRUE
		
		if (include_combined_matrix) {
			cmat = combine_matrices(matrices, mode="median")
			matrices = c(matrices, list(cmat))
		}
		
		# conditions should be a list of vectors of indices that map to the ratios input
		for (cond in conditions) {
			result = evaluate_matrices_ode(matrices, observed=ratios[,cond], counts=counts, init=ratios[,cond[1]],
				times=times, observation_indices=observation_indices,
				method=method, eval.mode=eval.mode, ave_results=TRUE, return.results=return.results, tau=tau)
			if (first) {
				first = FALSE
				results = result	
			} else {					
				results = cbind(results, result)
			}
		}	
		return(results)
}


evaluate_matrices_inferelator <- function(matrices, ratios, clusterStack, colMap, counts=NA, return.results=FALSE) {
	results = c()	
	predictions = list()
	for (m in matrices) {
		model = matrix_to_model(m)
		res = evaluate_predictive_model2(model, ratios, clusterStack, colMap)
		
		if (return.results) {
			predictions = c(predictions, res)	
		}
		
		performance = correlation_by_targets(res, counts=counts, return.asmean=TRUE)
		results = c(results, performance)
	}
	if (return.results) {
		return(predictions)	
	}
	return(results)
}

evaluate_matrices_ode <- function(matrices, ratios=NA, clusterStack=NA, colMap=NA, 
									observed=NA, counts=NA, init=NA, times=NA, observation_indices=NA, 
									method="ode45", eval.mode="correlation", tau=NA,
									ave_results=FALSE, return.results=FALSE) {
	# times have to match the number of times in the observed
	if (is.na(observed)) {
		results = evaluate_matrices_inferelator(list(matrices[[1]]), ratios, clusterStack, colMap, counts, return.results=TRUE)
		observed = results$observed
	}
	results = c()
	if (ave_results) {
		avr = matrix(0, length(matrices), length(times))
	}
	i = 0
	for (m in matrices) {
		ode_results = run_inferelator_ode(m, init=init, times=times, method=method, tau=tau)
		pred = t(ode_results)
		# ode simulation includes a 'time' row that we need to get rid of
		# and filter predictions for the observed times using observation_indices
		pred = pred[2:nrow(pred),]
		
		if (ave_results) {
			i = i + 1
			
			# normalize 0 to 1
			#pred = sapply(pred, function (a) (a-min(min(pred))/(max(pred)-min(pred))))
			#cat(pred, "\n")
			# this doesn't work- the pred has all the clusters
			avr[i,] = pred[1,]
		}
		#cat(colnames(observed), "\n")
		if (length(colnames(observed))==dim(pred)[[2]]) {
			colnames(pred) = colnames(observed)
		}
		prediction = list(observed=observed, predicted=pred)
		r = NA
		
		if (eval.mode == "correlation") {
			r = correlation_by_targets(prediction, counts=counts, return.asmean=TRUE)
		} else if (eval.mode == "rmsd") {
			r = rmsd_by_targets(prediction, counts=counts, return.asmean=TRUE, normalized=T)
		}
		if (return.results) {
			results = c(results, list(prediction))
		} else {
			results = c(results, r)
		}
	}
	
	if (ave_results) {
		avr = sapply(1:ncol(avr), function (r) mean(avr[,r]))
		
		avr = matrix(avr, nrow=1)
		colnames(avr) = colnames(observed)
		
		prediction = list(observed=observed, predicted=avr)
		if (eval.mode == "correlation") {
			r = correlation_by_targets(prediction, counts=counts, return.asmean=TRUE)
		} else {
			r = rmsd_by_targets(prediction, counts=counts, return.asmean=TRUE)
		}
		if (return.results) {
			results = c(results, list(prediction))
		} else {
			results = c(results, r)
		}
	}
	
	return(results)
}

plot_prediction <- function(predstruct, timesobs, timespred, cluster=1, title=NA, 
							xlab="Time (h)", ylab="Fold-change (log2) observed expression") {
	# function to plot predicted versus observed expression for a prediction when the
	#     times don't match- scales the predicted expression to match the observed range
	pred = predstruct[[1]]$predicted[cluster,]
	obsv = predstruct[[1]]$observed[cluster,]
		
	plot(list(x=timesobs, y=obsv), type="l", xlab=xlab, ylab=ylab, main=title)
	par(new=T)
	plot(list(x=timespred, y=pred),,type="l",col="green",xaxt="n",yaxt="n",xlab="",ylab="")
}

plot_predictions <- function(predstructlist, timesobs, timespred, predlabels, clusters=c(1)) {
	# does the same for a list of predictions
	x = length(predstructlist)
	y = length(clusters)
	par(mfrow=c(y,x))
	par(mar=c(3,3,2,1))
	
	for (cluster in clusters) {
		i = 1
		first = TRUE
		for (pred in predstructlist) {
			ylab = ""
			par(mar=c(3,3,2,1))
			if (first == TRUE) {
				par(mar=c(3,5,2,1))
				ylab = paste("Cluster ", cluster, sep="")
				first = FALSE
			}
			plot_prediction(pred, timesobs, timespred, cluster, title=predlabels[i], ylab=ylab)
			i = i + 1
		}
	}	
}

solver_example <- function() {
	func <- function(t, y, p) {
	    S <- y[1:2]	
	    I <- y[3:4]
	    R <- y[5:6]
	    dS <-  - p$beta * S * I
	    dI <-    p$beta * S *I - p$nu * I	
	    dR <-    p$nu * I
	    list(c(dS, dI, dR))
	}

	## note that parms can also be a list
	parms <- list(beta=c(0.1, 0.2), nu=c(0.1, 0.1))
	init <- c(S1 = 10, S2 = 8, I1 = 1, 
			I2 = 2, R1 = 0, R2 = 0)
	times <- seq(1, 10, 0.1)

	out <- ode(init, times, func, parms)

	plot(out)
}

inferelator_ode_function <- function(timenow, values, params, infmatrix, tau=1) {
	# this function is used by ode solver to determine the derivatives
	#      of the model at time 'timenow'- has to return a list with the first
	#      item being a vector of derivatives
	
	
	# Assumptions that need to be verified:
	#    the order of values is the same order as the
	#        input init structure, and thus the same order
	#        as the matrix we'll be using
	#    values have the same names as the init- thus the cols/rows
	#        of the matrix
	derivs = c()
	for (row in rownames(infmatrix)) {
		# each row is a target so calculate a new value (I think)
		weight = 0
		for (col in colnames(infmatrix)) {
			if (! is.na(infmatrix[row,col])) {
				weight = weight + (values[col]*infmatrix[row,col])
			}
		}
		#dydt = -values[col] + 1/(1+exp(1)^-weight)
		#cat(row, "\t", weight, "\t", dydt, "\n")
		#derivs = c(derivs, dydt)
		
		dydt = weight/tau
		
		derivs = c(derivs, dydt)

	}
	results = list(c(derivs))
	return(results)
}

run_inferelator_ode <- function(model_matrix, init, times, tau=15, method="ode45") {
	params = c()
	method = "euler"
	out = ode(init, times, inferelator_ode_function, params, infmatrix=model_matrix, method=method, tau=tau, hini=1, atol=1e-4, rtol=1e-4)
	return(out)
}

inferelator_ode_explicit_solver <- function(model, times, initial, tau, time0=NA) {
	# NOTE: this is not a correct solution for all models (and doesn't seem to work)
	#       I'm leaving it here for reference.
	# mathematical solution to our models
	# - model is the matrix of influences where diagonal values are -1
	# - times is a vector of times to produce as output
	# - time0 is the time at which initial was measured (defaults to first entry in times)
	# - initial is a vector of initial values (same number of rows and same order as model)
	# - tau is the timestep to use for the simulation
	
	start_t = 1
	end_t = length(times)
	collabel = c(time0, times)
	
	if (is.na(time0)) {
		time0 = times[1]
		start_t = 2
		collabel = times
	}
	
	result = matrix(initial, ncol=1)
	
	for (t in times[start_t:end_t]) {
		cmat = exp(((t-time0)/tau)*model)
		cvec = sapply(1:nrow(cmat), function (r) mean(cmat[r,]))
		c_at_t = cvec*initial
		result = cbind(result, c_at_t)
	}
	
	colnames(result) = collabel
	return(result)
}


correlation_by_targets <- function(results, minc=1, condrange=NA, dynamic=FALSE, counts=NA, return.asmean=FALSE) {
	if (is.na(condrange)) {
		condrange = 1:ncol(results$predicted)
		#condrange = 1:length(colnames(results$predicted))
		}
	
	if (dynamic) {
		m = sapply(1:dim(results$dynamic)[1], function(i) cor(results$dynamic[i,condrange], results$observed[i,condrange], method="spearman", use="complete.obs"))	
	} else {
		m = c()
		for (i in 1:dim(results$predicted)[1]) {
			r = try(cor(results$predicted[i,condrange], results$observed[i,condrange], method="pearson", use="complete.obs"), silent=T)
			if (class(r) == "try-error") {
				r = NA
			}
			m = c(m,r)
		}
		#m = sapply(1:dim(results$predicted)[1], function(i) cor(results$predicted[i,condrange], results$observed[i,condrange], method="pearson", use="complete.obs"))
	}
	mn = mean(m, na.rm=TRUE)
	if (! is.na(counts)) {
		mn = sum(sapply(1:length(m), function(i) m[i]*counts[i]), na.rm=TRUE)/sum(counts)
	}
	if (return.asmean) {
		return(mn)	
	}
	cat("Mean correlation of targets (pY to Y): ", mn, "\n")
	names(m) = rownames(results$predicted)
	return(as.vector(m))
}


rmsd_by_targets <- function(results, maxc=0, dynamic=FALSE, counts=NA, normalized=TRUE, return.asmean=FALSE) {
	if (maxc == 0) {
		maxc = length(results$predicted[,1])	
	}
	
	if (dynamic) {
		m = sapply(1:dim(results$dynamic[1:maxc,])[1], function(i) sqrt(sum((results$observed[i,]-results$dynamic[i,])^2)/dim(results$dynamic)[2]))	
	} else {
		if (normalized) {
			m = sapply(1:dim(results$predicted[1:maxc,])[1], function(i) (sqrt(sum((results$observed[i,]-results$predicted[i,])^2)/dim(results$predicted)[2]))/(max(results$observed[i,]-min(results$observed[i,]))))	
		} else {
		# wow, is this convoluted enough?
			m = sapply(1:dim(results$predicted[1:maxc,])[1], function(i) sqrt(sum((results$observed[i,]-results$predicted[i,])^2)/dim(results$predicted)[2]))
		}
	}
	mn = mean(m)
	if (! is.na(counts)) {
		mn = sum(sapply(1:length(m), function(i) m[i]*counts[i]))/sum(counts)
	}
	if (return.asmean) {
		return(mn)	
	}
	cat("Mean RMSD of targets (pY to Y):", mn, "\n")
	names(m) = rownames(results$predicted[1:maxc,])
	return(m)
}

